package net.wanji.com.netty;


import io.netty.bootstrap.ServerBootstrap;
import io.netty.channel.*;
import io.netty.channel.group.ChannelGroup;
import io.netty.channel.group.DefaultChannelGroup;
import io.netty.channel.nio.NioEventLoopGroup;
import io.netty.channel.socket.nio.NioServerSocketChannel;
import io.netty.channel.socket.nio.NioSocketChannel;
import io.netty.handler.codec.LengthFieldBasedFrameDecoder;
import io.netty.util.concurrent.GlobalEventExecutor;
import lombok.extern.slf4j.Slf4j;
import net.wanji.com.cache.netty.NettyMessageCache;
import net.wanji.com.netty.handler.NettyServerHandler;
import net.wanji.com.pojo.netty.MessageResultPojo;
import org.apache.commons.lang3.StringUtils;
import org.springframework.scheduling.annotation.Scheduled;
import org.springframework.scheduling.concurrent.ThreadPoolTaskExecutor;
import org.springframework.stereotype.Component;

import javax.annotation.Resource;
import java.nio.ByteOrder;
import java.time.LocalDateTime;
import java.time.ZoneOffset;
import java.util.HashMap;
import java.util.Map;
import java.util.Objects;
import java.util.concurrent.CountDownLatch;

@Slf4j
@Component
public class NettyServer {

    private static final Map<String, ChannelId> IP_PORT_CHANNEL_ID_MAP = new HashMap<>();
    private static ChannelGroup channelGroup = new DefaultChannelGroup(GlobalEventExecutor.INSTANCE);
    @Resource(name = "threadPoolExecutor")
    ThreadPoolTaskExecutor threadPoolExecutor;

    /**
     * 建立连接
     */
    public static void start(int port) throws InterruptedException {
        final EventLoopGroup parent = new NioEventLoopGroup();
        final EventLoopGroup child = new NioEventLoopGroup();
        ServerBootstrap serverBootstrap = new ServerBootstrap();
        serverBootstrap.group(parent, child)
                .channel(NioServerSocketChannel.class)
                .childOption(ChannelOption.TCP_NODELAY, true)
                .childOption(ChannelOption.SO_KEEPALIVE, true)
                .childHandler(new ChannelInitializer<NioSocketChannel>() {
                    @Override
                    protected void initChannel(NioSocketChannel socketChannel) {
                        socketChannel.pipeline()
                                .addFirst()
//                                .addLast(new MessageDecoder())
//                                .addLast(new MessageEnCoder())
                                // 格林威解码器
                                .addLast(new LengthFieldBasedFrameDecoder(ByteOrder.LITTLE_ENDIAN, Integer.MAX_VALUE, 4, 2, 0, 0, true))
                                .addLast(new NettyServerHandler());
                        channelGroup.add(socketChannel);
                        String key = StringUtils.join(socketChannel.remoteAddress().getHostString(), ":".intern(), socketChannel.remoteAddress().getPort());
                        log.info("client {} is connect success", key);
                        IP_PORT_CHANNEL_ID_MAP.put(key, socketChannel.id());
                    }
                });

        ChannelFuture channelFuture = serverBootstrap.bind(port).sync();
        log.info("server is started in {}", channelFuture.channel().localAddress());
        //收到停止信号
        Runtime.getRuntime().addShutdownHook(new Thread(() -> {
            //停掉所有的线程
            channelGroup.close().syncUninterruptibly();
            parent.shutdownGracefully();
            child.shutdownGracefully();
            channelFuture.channel().closeFuture().syncUninterruptibly();
            log.warn("server is closed");
        }));
    }

    public static void sendMessage(String ip, int port, String message) {
        Channel channel = channelGroup.find(IP_PORT_CHANNEL_ID_MAP.get(StringUtils.join(ip, ":", port)));
        channel.writeAndFlush(message);
    }

    public static Object sendMessage(String ip, int port, String message, String command) {
        sendMessage(ip, port, message, command, 500);
        return null;
    }

    public static String sendMessage(String ip, int port, String message, String command, int waitMillisecond) {
        CountDownLatch countDownLatch = new CountDownLatch(1);
        String key = StringUtils.joinWith("#", ip, port, command);
        long now = LocalDateTime.now().toInstant(ZoneOffset.of("+8")).toEpochMilli();
        MessageResultPojo messageResultPojo = new MessageResultPojo(command, countDownLatch, now, waitMillisecond, null);
        NettyMessageCache.NETTY_MESSAGE_RESULT_MAP.put(key, messageResultPojo);
        try {
            sendMessage(ip, port, message);
            countDownLatch.await();
        } catch (InterruptedException e) {
            log.error("error", e);
        }
        MessageResultPojo resultPojo = NettyMessageCache.NETTY_MESSAGE_RESULT_MAP.remove(key);
        if (Objects.nonNull(resultPojo)) {
            return resultPojo.getHexMessageResult();
        }
        return null;
    }

    @Scheduled(fixedRate = 300)
    public void checkWaitTimeout() {
        long now = LocalDateTime.now().toInstant(ZoneOffset.of("+8")).toEpochMilli();
        NettyMessageCache.NETTY_MESSAGE_RESULT_MAP.forEach((k, v) -> {
            threadPoolExecutor.execute(() -> {
                if (now - v.getStartTime() > v.getWaitMillisecond()) {
                    log.warn("wait {}ms {} timeout", v.getWaitMillisecond(), k);
                    v.getCountDownLatch().countDown();
                }
            });
        });
    }
}